import numpy as np
import copy
import pandas as pd
from soft_impute_test import SoftImpute
# from soft_impute import SoftImpute
import os, sys
import matplotlib.pyplot as plt
from scipy.stats import entropy
import logging
import random
from sklearn.metrics import r2_score, f1_score
import warnings
warnings.filterwarnings("ignore")


def Generate_binary_result(prob):
    rand = np.random.uniform()
    if rand < prob:
        return True
    else: return False

def Preparing_Labels(labelnums):
    labels = []
    for i in range(0, labelnums):
        labels.append(i)
    return labels

def matrix_factorization(R, P, Q, K, steps=5000, alpha=0.0002, beta=0.02):
    Q = Q.T
    for step in range(steps):
        for i in range(len(R)):
            for j in range(len(R[i])):
                if R[i][j] != -1:
                    eij = R[i][j] - np.dot(P[i,:],Q[:,j])
                    for k in range(K):
                        P[i][k] = P[i][k] + alpha * (2 * eij * Q[k][j] - beta * P[i][k])
                        Q[k][j] = Q[k][j] + alpha * (2 * eij * P[i][k] - beta * Q[k][j])
        eR = np.dot(P,Q)
        e = 0
        for i in range(len(R)):
            for j in range(len(R[i])):
                if R[i][j] > 0:
                    e = e + pow(R[i][j] - np.dot(P[i,:],Q[:,j]), 2)
                    for k in range(K):
                        e = e + (beta/2) * (pow(P[i][k],2) + pow(Q[k][j],2))
        if e < 0.001:
            break
    return P, Q.T


def Generating_Truth(num, uniqueness, responsiveness, numtypes):
    labels = list(range(numtypes))
    table = []
    unique_num = int(num * uniqueness)
    for i in range(unique_num):
        row = []
        rand = np.random.randint(len(labels))
        row.append(labels[rand])
        table.append(row)
    for i in range(unique_num):
        rest_pool = labels.copy()
        rest_pool.remove(table[i][0])
        for j in range(unique_num - 1):
            if Generate_binary_result(responsiveness):
                table[i].append(rest_pool[np.random.randint(len(rest_pool))])
            else:
                table[i].append(table[i][0])
    for j in range(num - unique_num):
        rand = np.random.randint(unique_num)
        for i in range(unique_num):
            table[i].append(table[i][rand])
    for i in range(num - unique_num):
        table.append(table[np.random.randint(unique_num)])
    table = np.array(table)
    #table = cp.array(table)
    filename = str(uniqueness) + 'test_data.csv'
    # pd_data = pd.DataFrame(table)
    # pd_data.to_csv('test_data.csv',index=False,header=False)
    return table


def Initialize_Data(ground_truth, initial_num):
    currdata = []
    for i in range(len(ground_truth)):
        row = []
        for j in range(len(ground_truth[i])):
            row.append(-1)
        currdata.append(row)

    for i in range(initial_num):
        rand1 = np.random.randint(len(ground_truth))
        rand2 = np.random.randint(len(ground_truth[0]))
        while currdata[rand1][rand2] != -1 or ground_truth[rand1][rand2] == -1:
            rand1 = np.random.randint(len(ground_truth))
            rand2 = np.random.randint(len(ground_truth[0]))
        currdata[rand1][rand2] = ground_truth[rand1][rand2]
    return currdata

def CountKnownNum(currdata):
    count = 0
    for row in currdata:
        for e in row:
            if e != -1:
                count+=1
    return count

def zeros_like(m):
    mtx = []
    for i in range(len(m)):
        row = []
        for j in range(len(m[i])):
            row.append(0)
        mtx.append(row)
    return mtx

class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout

def Make_Selections_score2(num, predictions):
    batch = []
    for prediction in predictions:
        if len(batch) < num:
            batch.append(prediction)  #prediction format (row, col, label, score)
        else:
            max_score = batch[0][3]
            index = 0
            for i in range(len(batch)):
                if batch[i][3] > max_score:
                    max_score = batch[i][3]
                    index = i

            if prediction[3] < max_score:
                batch.remove(batch[index])
                batch.append(prediction)
    return batch

def Make_Selections_score(num, predictions):
    batch = []
    for prediction in predictions:
        if len(batch) < num:
            batch.append(prediction)  #prediction format (row, col, label, score)
        else:
            min_score = batch[0][3]
            index = 0
            for i in range(len(batch)):
                if batch[i][3] < min_score:
                    min_score = batch[i][3]
                    index = i
            if prediction[3] > min_score:
                batch.remove(batch[index])
                batch.append(prediction)
    return batch
    
def Make_Selections_random(num, predictions):
    batch = []
    batch_index = []
    for i in range(num):
        rand = np.random.randint(len(predictions))
        while rand in batch_index:
            rand = np.random.randint(len(predictions))
        batch_index.append(rand)
    for index in batch_index:
        batch.append(predictions[index])
    return batch


def svt_solve(A, mask, tau=None, delta=None, epsilon=1e-2, max_iterations=1000):
  """
  Solve using iterative singular value thresholding.
  [ Cai, Candes, and Shen 2010 ]
  Parameters:
  -----------
  A : m x n array
    matrix to complete
  mask : m x n array
    matrix with entries zero (if missing) or one (if present)
  tau : float
    singular value thresholding amount;, default to 5 * (m + n) / 2
  delta : float
    step size per iteration; default to 1.2 times the undersampling ratio
  epsilon : float
    convergence condition on the relative reconstruction error
  max_iterations: int
    hard limit on maximum number of iterations
  Returns:
  --------
  X: m x n array
    completed matrix
  """
  logger = logging.getLogger(__name__)
  Y = np.zeros_like(A)

  if not tau:
    tau = 5 * np.sum(A.shape) / 2
  if not delta:
    delta = 1.2 * np.prod(A.shape) / np.sum(mask)

  for _ in range(max_iterations):

    U, S, V = np.linalg.svd(Y, full_matrices=False)

    S = np.maximum(S - tau, 0)

    X = np.dot(np.dot(U, np.diag(S)), V)
    Y += delta * mask * (A - X)

    rel_recon_error = np.linalg.norm(mask * (X - A)) / np.linalg.norm(mask * A)
    if _ % 1 == 0:
      logger.info("Iteration: %i; Rel error: %.4f" % (_ + 1, rel_recon_error))
    if rel_recon_error < epsilon:
      break
  return X

def main(uniqueness, responsiveness, query):
    phenotype = 17
    ground_truth = np.load("experimental-space-v12.npy")
    current_mtx = Initialize_Data(ground_truth, 400)
    total = len(current_mtx) * len(current_mtx[0])
    total0 = CountKnownNum(ground_truth)
    affl_matrix = []
    masks = []
    knownnum = CountKnownNum(current_mtx)
    accuracy = []
    pred_accuracy = []
    batch_size = 400
    
    f1weighted = []
    f1unweighted = []
    while knownnum < total0:
        current_phenotypes = np.unique(current_mtx)
        affl_matrix = []
        del_idx = []
        for i in range(len(current_phenotypes)):
            if current_phenotypes[i] < 0:
                del_idx.append(i)
        current_phenotypes = np.delete(current_phenotypes, del_idx)
        mapping = {}
        mapping_inv = {}
        for i in range(len(current_phenotypes)):
            if current_phenotypes[i] >= 0:
                mapping[i] = current_phenotypes[i]
                mapping_inv[current_phenotypes[i]] = i
        
        for k in range(len(mapping)):
            mtx = np.zeros_like(current_mtx).astype('f')
            affl_matrix.append(mtx)
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if current_mtx[i][j] != -1:
                    idx = mapping_inv[current_mtx[i][j]]
                    for k in range(len(mapping)):
                        if k == idx:
                            affl_matrix[k][i][j] = 1
                else: 
                    for k in range(len(mapping)):
                        affl_matrix[k][i][j] = 'NaN'
        prediction = []
        with HiddenPrints():
            for k in range(len(mapping)):
                # solver = SoftImpute(J=2, maxit=1000).fit(affl_matrix[k])
                # affl_matrix[k] = SoftImpute(max_iters=500).fit_transform(affl_matrix[k])
                affl_matrix[k] = SoftImpute(max_iters=100, init_fill_method="half", verbose=False).fit_transform(affl_matrix[k])

        pred4f1 = []
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if ground_truth[i, j] != -1:
                    if current_mtx[i][j] == -1:
                        temp = []
                        for k in range(len(mapping)):
                            temp.append(affl_matrix[k][i][j])
                        label = mapping[np.argmax(temp)]
                        temp.sort(reverse=True)
                        pred4f1.append([label, ground_truth[i, j]])
                        if query == 'active': 
                            score = temp[0] - temp[1]
                            prediction.append([i, j, label, score])
                        elif query == 'least_confidence':
                            score = temp[0]
                            prediction.append([i, j, label, score])
                        elif query == 'entropy' or query == 'random':
                            temp2 = []
                            for value in temp:
                                temp2.append(value)
                            temp2 /= np.sum(temp2)
                            score = entropy(temp2, base=2)
                            prediction.append([i, j, label, score])
                        else: raise ValueError('unknown')
                    else:
                        pred4f1.append([ground_truth[i, j], ground_truth[i, j]])   #known
        pred4f1 = np.array(pred4f1)
        f1weighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='weighted'))
        f1unweighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='macro'))     
        size = min(batch_size, total0 - knownnum)
        if query == 'active': 
            batch = Make_Selections_score2(size, prediction)
        elif query == 'least_confidence':
            batch = Make_Selections_score2(1000, prediction)
        elif query == 'entropy':
            batch = Make_Selections_score(1000, prediction)
        elif query == 'random':
            batch = Make_Selections_random(size, prediction)
        else: raise ValueError('unknown')
        affl_matrix.clear()
        masks.clear()
        mistake = 0
        for p in prediction:
            i = p[0]
            j = p[1]
            label = p[2]
            if ground_truth[i][j] != label:
                mistake += 1
        accuracy.append((total0 - mistake)/total0)
        pred_accuracy.append(1-mistake / len(prediction))
        for item in batch:
            i = item[0]
            j = item[1]
            current_mtx[i][j] = ground_truth[i][j]
        knownnum = CountKnownNum(current_mtx)
    accuracy.append(1.0)
    f1unweighted.append(1.0)
    f1weighted.append(1.0)
    filename = 'si_tc_' + str(query) + '_acc.csv'
    result = []
    result.append(accuracy)
    # result.append(pred_accuracy)
    pd_data = pd.DataFrame(result)
    pd_data.to_csv(filename,index=False,header=False, mode="a")
    
    result1 = []
    result1.append(f1unweighted)
    pd_data = pd.DataFrame(result1)
    pd_data.to_csv("si_tc_" + str(query) + "_unwf1.csv", index=False, header=False, mode="a")
    result2 = []
    result2.append(f1weighted)
    pd_data = pd.DataFrame(result2)
    pd_data.to_csv("si_tc_" + str(query) + "_wf1.csv", index=False, header=False, mode="a")
np.random.seed(10)
random.seed(10)


main(0.4, 0.2, 'active')
main(0.4, 0.2, 'random')



