import matlab.engine
import numpy as np
from numpy.linalg import matrix_rank
from scipy.stats import entropy
import sys, os
import pandas as pd
import matplotlib.pyplot as plt
from doc6b import *
import random
import copy
from sklearn.metrics import f1_score


def Generate_binary_result(prob):
    rand = np.random.uniform()
    if rand < prob:
        return True
    else: return False

def Generating_Truth(num, uniqueness, responsiveness, numtypes):
    labels = list(range(0, numtypes))
    table = []
    unique_num = int(num * uniqueness)
    for i in range(unique_num):
        row = []
        rand = np.random.randint(len(labels))
        row.append(labels[rand])
        table.append(row)
    for i in range(unique_num):
        rest_pool = labels.copy()
        rest_pool.remove(table[i][0])
        for j in range(unique_num - 1):
            if Generate_binary_result(responsiveness):
                table[i].append(rest_pool[np.random.randint(len(rest_pool))])
            else:
                table[i].append(table[i][0])
    for j in range(num - unique_num):
        rand = np.random.randint(unique_num)
        for i in range(unique_num):
            table[i].append(table[i][rand])
    for i in range(num - unique_num):
        table.append(table[np.random.randint(unique_num)])
    table = np.array(table)
    #table = cp.array(table)
    filename = str(uniqueness) + 'test_data.csv'
    # pd_data = pd.DataFrame(table)
    # pd_data.to_csv('test_data.csv',index=False,header=False)
    return table


def Initialize_Data(ground_truth, num):
    currdata = []
    for i in range(len(ground_truth)):
        row = []
        for j in range(len(ground_truth[i])):
            row.append(-1)
        currdata.append(row)

    for i in range(num):
        rand1 = np.random.randint(len(ground_truth))
        rand2 = np.random.randint(len(ground_truth[0]))
        while currdata[rand1][rand2] != -1:
            rand1 = np.random.randint(len(ground_truth))
            rand2 = np.random.randint(len(ground_truth[0]))
        currdata[rand1][rand2] = ground_truth[rand1][rand2]
    return currdata

def one_bit_solve(mtx_, eng, idx_):
    mtx_ = np.array(mtx_)
    y_ = list(mtx_.T.reshape(-1,))
    y = []
    for v in y_:
        y.append(int(v))
    r = float(20)
    recovered = eng.solver(idx_, y, r, float(94), float(92))
    return np.asarray(recovered)


def CountKnownNum(currdata):
    count = 0
    for row in currdata:
        for e in row:
            if e != -1:
                count+=1
    return count

def neg_like(m):
    mtx = []
    for i in range(len(m)):
        row = []
        for j in range(len(m[i])):
            row.append(-1)
        mtx.append(row)
    return mtx

class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout
def Make_Selections_score(num, predictions):
    prediction_ = copy.deepcopy(predictions)
    random.shuffle(prediction_)
    sorted_list = sorted(prediction_, key=lambda x : x[3], reverse = True)
    batch = sorted_list[: int(num)]
    return batch


def Make_Selections_score2(num, predictions):
    prediction_ = copy.deepcopy(predictions)
    random.shuffle(prediction_)
    sorted_list = sorted(prediction_, key=lambda x : x[3], reverse = False)
    batch = sorted_list[: int(num)]
    return batch



def make_selections_random(num, predictions):
    batch = []
    batch_index = []
    for i in range(num):
        rand = np.random.randint(len(predictions))
        while rand in batch_index:
            rand = np.random.randint(len(predictions))
        batch_index.append(rand)
    for index in batch_index:
        batch.append(predictions[index])
    return batch


def main(query):
    eng = matlab.engine.start_matlab()
    data = np.array(Readin_xls('elife-10047-supp4-v2.xls'))
    init_idx = np.random.choice(range(4*len(data)), int(0.04*len(data)), replace=False)
    measured_samples = []
    new_added = []
    measured_idx = {}
    init_instances = []
    for i in init_idx:
        row = int(i / 92)
        col = i % 92
        init_instances.append((row, col))
        if row > 46:
            row = row - 47
        if col > 45:
            col = col - 46
        new_added.append((row, col))

    for sample in data:
        row = sample[1]-1
        col = sample[0]-1
        if (row, col) in new_added and  (row, col) not in measured_idx:
            measured_samples.append(sample)
    for i in new_added:
        row = i[0]
        col = i[1]
        if (row, col) in measured_idx: 
            measured_idx[(row, col)] += 1
        else: measured_idx[(row, col)] = 1

   
    measured_samples = np.array(measured_samples)
    clustering = AgglomerativeClustering(distance_threshold=10, n_clusters=None).fit(measured_samples[:, 4:])
    measured_y = clustering.labels_

    knn = KNeighborsClassifier(n_neighbors = 5, weights = 'distance').fit(measured_samples[:, 4:], measured_y)
    ground_truth = Construct_Groundtruth(data, 47, 46, knn)
    current_mtx = -np.ones((94, 92))
    for i in init_instances:
        current_mtx[i[0], i[1]] = ground_truth[i[0], i[1]]
    

    total = len(current_mtx) * len(current_mtx[0])
    affl_matrix = []
    knownnum = CountKnownNum(current_mtx)
    accuracy = []
    pred_accuracy = []
    vectors = []
    f1weighted = []
    f1unweighted = []


    once = 0
    twice = 0
    thriple = 0
    quartic = 0
    for key in measured_idx:
        if measured_idx[key] == 1:
            once += 1
        elif measured_idx[key] == 2:
            twice += 1
        elif measured_idx[key] == 3:
            thriple += 1
        elif measured_idx[key] == 4:
            quartic += 1
    zeros = total/4 - once - twice - thriple - quartic
    vector = (zeros*4/total, once*4/total, twice*4/total, thriple*4/total, quartic*4/total)
    vectors.append(vector)
    while knownnum < total:
        phenotype = len(np.unique(measured_y))
        idx_= []
        for p in range(92):
            for q in range(94):
                if current_mtx[q][p] != -1:
                    idx_.append(q+1+p*94)
        for k in range(phenotype):
            mtx = -np.ones_like(current_mtx)
            affl_matrix.append(mtx)
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if current_mtx[i][j] != -1:
                    idx = int(ground_truth[i][j])
                    for k in range(phenotype):
                        if k == idx:
                            affl_matrix[k][i][j] = 1
                else:
                    for k in range(phenotype):
                        affl_matrix[k][i][j] = -1
 
        prediction = []
        with HiddenPrints():
            for k in range(phenotype):
                # solver = SoftImpute(J=2, maxit=1000).fit(affl_matrix[k])
                # affl_matrix[k] = SoftImpute(max_iters=500).fit_transform(affl_matrix[k])
                affl_matrix[k] = one_bit_solve(affl_matrix[k], eng, idx_)

        pred4f1 = []
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if ground_truth[i, j] != -1:
                    temp = []
                    for k in range(phenotype):
                        temp.append(affl_matrix[k][i][j])
                    label = np.argmax(temp)
                    temp.sort(reverse=True)
                    if current_mtx[i][j] == -1:
                        pred4f1.append([label, ground_truth[i, j]])
                    else:
                        pred4f1.append([ground_truth[i, j], ground_truth[i, j]])   #known
                    if current_mtx[i][j] == -1:
                        if query == 'margin': 
                            score = temp[0] - temp[1]
                            prediction.append([i, j, label, score])
                        elif query == 'active' or query == 'random':
                            score = temp[0]
                            prediction.append([i, j, label, score])
                        elif query == 'entropy':
                            temp2 = []
                            for value in temp:
                                temp2.append(value)
                            temp2 /= np.sum(temp2)
                            score = entropy(temp2, base=2)
                            prediction.append([i, j, label, score])
                        else: raise ValueError('unknown')

        pred4f1 = np.array(pred4f1)
        f1weighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='weighted'))
        f1unweighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='macro'))
        size = min(int(0.01*total), total-knownnum)
        if query == 'margin': 
            batch = Make_Selections_score2(000, prediction)
        elif query == 'active':
            batch = Make_Selections_score2(size, prediction)
        elif query == 'entropy':
            batch = Make_Selections_score(1000, prediction)
        elif query == 'random':
            batch = make_selections_random(size, prediction)
        else: raise ValueError('unknown')
        affl_matrix.clear()
        mistake = 0
        for p in prediction:
            i = p[0]
            j = p[1]
            label = p[2]
            if ground_truth[i][j] != label:
                mistake += 1
        accuracy.append((94*92 - mistake)/(94*92))
        pred_accuracy.append((len(prediction)-mistake)/len(prediction))
        new_added = []
        for i in range(len(batch)):
            current_mtx[batch[i][0]][batch[i][1]] = ground_truth[batch[i][0]][batch[i][1]]
            if batch[i][0] > 46:
                row = batch[i][0] - 47
            else: row = batch[i][0]
            if batch[i][1] > 45:
                col = batch[i][1] - 46
            else: col = batch[i][1]
            new_added.append((row, col))


        
        for sample in data:
            row = sample[1]-1
            col = sample[0]-1
            if (row, col) in new_added and (row, col) not in measured_idx:
                measured_samples = np.vstack((measured_samples, sample))
        for i in new_added:
            row = i[0]
            col = i[1]
            if (row, col) in measured_idx: 
                measured_idx[(row, col)] += 1
            else: measured_idx[(row, col)] = 1


        knownnum = CountKnownNum(current_mtx)

        once = 0
        twice = 0
        thriple = 0
        quartic = 0
        for key in measured_idx:
            if measured_idx[key] == 1:
                once += 1
            elif measured_idx[key] == 2:
                twice += 1
            elif measured_idx[key] == 3:
                thriple += 1
            elif measured_idx[key] == 4:
                quartic += 1
        zeros = total/4 - once - twice - thriple - quartic
        vector = (zeros*4/total, once*4/total, twice*4/total, thriple*4/total, quartic*4/total)
        vectors.append(vector)

        clustering = AgglomerativeClustering(distance_threshold=10, n_clusters=None).fit(measured_samples[:, 4:])
        #clustering = AgglomerativeClustering(n_clusters=60).fit(measured_samples[:, 4:])


        measured_y = clustering.labels_
        knn = KNeighborsClassifier(n_neighbors = 5, weights = 'distance').fit(measured_samples[:, 4:], measured_y)
        ground_truth = Construct_Groundtruth(data, 47, 46, knn)
        current_mtx = Update_currdata(current_mtx, ground_truth)
    filename = 'onebit_image_' + str(query) + '_acc.csv'
    result = []
    result.append(accuracy)
    # result.append(pred_accuracy)
    pd_data = pd.DataFrame(result)
    pd_data.to_csv(filename,index=False,header=False,mode="a")

    result1 = []
    result1.append(f1unweighted)
    pd_data = pd.DataFrame(result1)
    pd_data.to_csv("onebit_image_" + str(query) + "_unwf1.csv", index=False, header=False, mode="a")
    result2 = []
    result2.append(f1weighted)
    pd_data = pd.DataFrame(result2)
    pd_data.to_csv("onebit_image_" + str(query) + "_wf1.csv", index=False, header=False, mode="a")


random.seed(10)
np.random.seed(10)
main("active")
main("random")