import matlab.engine
import numpy as np
from numpy.linalg import matrix_rank
from scipy.stats import entropy
import sys, os
import pandas as pd
import matplotlib.pyplot as plt
import time
import random
from sklearn.metrics import f1_score


def Generate_binary_result(prob):
    rand = np.random.uniform()
    if rand < prob:
        return True
    else: return False

def Generating_Truth(num, uniqueness, responsiveness, numtypes):
    labels = list(range(0, numtypes))
    table = []
    unique_num = int(num * uniqueness)
    for i in range(unique_num):
        row = []
        rand = np.random.randint(len(labels))
        row.append(labels[rand])
        table.append(row)
    for i in range(unique_num):
        rest_pool = labels.copy()
        rest_pool.remove(table[i][0])
        for j in range(unique_num - 1):
            if Generate_binary_result(responsiveness):
                table[i].append(rest_pool[np.random.randint(len(rest_pool))])
            else:
                table[i].append(table[i][0])
    for j in range(num - unique_num):
        rand = np.random.randint(unique_num)
        for i in range(unique_num):
            table[i].append(table[i][rand])
    for i in range(num - unique_num):
        table.append(table[np.random.randint(unique_num)])
    table = np.array(table)
    #table = cp.array(table)
    filename = str(uniqueness) + 'test_data.csv'
    # pd_data = pd.DataFrame(table)
    # pd_data.to_csv('test_data.csv',index=False,header=False)
    return table


def Initialize_Data(ground_truth, num):
    currdata = []
    for i in range(len(ground_truth)):
        row = []
        for j in range(len(ground_truth[i])):
            row.append(-1)
        currdata.append(row)

    for i in range(num):
        rand1 = np.random.randint(len(ground_truth))
        rand2 = np.random.randint(len(ground_truth[0]))
        while currdata[rand1][rand2] != -1 or ground_truth[rand1][rand2] == -1:
            rand1 = np.random.randint(len(ground_truth))
            rand2 = np.random.randint(len(ground_truth[0]))
        currdata[rand1][rand2] = ground_truth[rand1][rand2]
    # print(currdata[0][0] == ground_truth[0][0])
    return currdata

def AddNoise(data, percent, phenotype):
    added = []
    noise_num = int(len(data) * len(data[0]) * percent)
    i = 0
    while i < noise_num:
        row = np.random.randint(0, len(data))
        col = np.random.randint(0, len(data[0]))
        while (row, col) in added:
             row = np.random.randint(0, len(data))
        col = np.random.randint(0, len(data[0]))
        noise_index = np.random.randint(0, phenotype)
        while data[row][col] == noise_index:
            noise_index = np.random.randint(0, phenotype)
        data[row][col] = noise_index
        added.append((row, col))
        i += 1
    return data

def one_bit_solve(mtx_, eng, idx_):
    mtx_ = np.array(mtx_)
    y_ = list(mtx_.T.reshape(-1,))
    y = []
    for v in y_:
        y.append(int(v))
    r = float(50)
    recovered = eng.solver(idx_, y, r, float(100), float(100))
    return np.asarray(recovered)


def CountKnownNum(currdata):
    count = 0
    for row in currdata:
        for e in row:
            if e != -1:
                count+=1
    return count

def neg_like(m):
    mtx = []
    for i in range(len(m)):
        row = []
        for j in range(len(m[i])):
            row.append(-1)
        mtx.append(row)
    return mtx

class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout

def Make_Selections_score2(num, predictions):
    batch = []
    for prediction in predictions:
        if len(batch) < num:
            batch.append(prediction)  #prediction format (row, col, label, score)
        else:
            max_score = batch[0][3]
            index = 0
            for i in range(len(batch)):
                if batch[i][3] > max_score:
                    max_score = batch[i][3]
                    index = i

            if prediction[3] < max_score:
                batch.remove(batch[index])
                batch.append(prediction)
    return batch

def Make_Selections_score(num, predictions):
    batch = []
    for prediction in predictions:
        if len(batch) < num:
            batch.append(prediction)  #prediction format (row, col, label, score)
        else:
            min_score = batch[0][3]
            index = 0
            for i in range(len(batch)):
                if batch[i][3] < min_score:
                    min_score = batch[i][3]
                    index = i
            if prediction[3] > min_score:
                batch.remove(batch[index])
                batch.append(prediction)
    return batch


def main(uniqueness, responsiveness, query, noise):
    phenotype = 32
    ground_truth = Generating_Truth(100, uniqueness, responsiveness, phenotype)
    ground_truth = AddNoise(ground_truth, noise, phenotype)
    percentages = []
    # for k in range(1, 5):
    #     percentages.append(np.sum(np.where(ground_truth==k, 1, 0))/(ground_truth.shape[0]*ground_truth.shape[0]))
    # plt.bar(range(1, 5), percentages)
    # plt.show()
    phenotypes = np.unique(ground_truth)
    # ground_truth = np.random.randint(2, size = (100, 100))

    current_mtx = Initialize_Data(ground_truth, 100)
    total = len(current_mtx) * len(current_mtx[0])
    total0 = CountKnownNum(ground_truth)
    affl_matrix = []
    knownnum = CountKnownNum(current_mtx)
    accuracy = []
    pred_accuracy = []
    batch_size = 100

    f1weighted = []
    f1unweighted = []

    while knownnum < total0:
        affl_matrix = []
        knownnum = CountKnownNum(current_mtx)
        current_phenotypes = np.unique(current_mtx)
        del_idx = []
        for i in range(len(current_phenotypes)):
            if current_phenotypes[i] < 0:
                del_idx.append(i)
        current_phenotypes = np.delete(current_phenotypes, del_idx)
        mapping = {}
        mapping_inv = {}
        for i in range(len(current_phenotypes)):
            if current_phenotypes[i] >= 0:
                mapping[i] = current_phenotypes[i]
                mapping_inv[current_phenotypes[i]] = i
        for k in range(len(mapping)):
            mtx = -np.ones_like(current_mtx)
            affl_matrix.append(mtx)

        idx_= []
        for p in range(ground_truth.shape[1]):
            for q in range(ground_truth.shape[0]):
                if current_mtx[q][p] != -1:
                    idx_.append(q+1+p*ground_truth.shape[0])
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if current_mtx[i][j] != -1:
                    idx = mapping_inv[current_mtx[i][j]]
                    for k in range(len(mapping)):
                        if k == idx:
                            affl_matrix[k][i][j] = 1


        prediction = []
        with HiddenPrints():
            for k in range(len(mapping)):
                # solver = SoftImpute(J=2, maxit=1000).fit(affl_matrix[k])
                # affl_matrix[k] = SoftImpute(max_iters=500).fit_transform(affl_matrix[k])
                affl_matrix[k] = one_bit_solve(affl_matrix[k], eng, idx_)

        pred4f1 = []
        for i in range(len(current_mtx)):
            for j in range(len(current_mtx[i])):
                if ground_truth[i, j] != -1:
                    temp = []
                    for k in range(len(mapping)):
                        temp.append(affl_matrix[k][i][j])
                    label = mapping[np.argmax(temp)]
                    temp.sort(reverse=True)
                    if current_mtx[i][j] == -1:
                        pred4f1.append([label, ground_truth[i, j]])
                    else:
                        pred4f1.append([ground_truth[i, j], ground_truth[i, j]])   #known
                    if current_mtx[i][j] == -1:
                        if query == 'margin': 
                            score = temp[0] - temp[1]
                            prediction.append([i, j, label, score])
                        elif query == 'active' or query == 'random':
                            score = temp[0]
                            prediction.append([i, j, label, score])
                        elif query == 'entropy':
                            temp2 = []
                            for value in temp:
                                temp2.append(value)
                            temp2 /= np.sum(temp2)
                            score = entropy(temp2, base=2)
                            prediction.append([i, j, label, score])
                        else: raise ValueError('unknown')

        pred4f1 = np.array(pred4f1)
        f1weighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='weighted'))
        f1unweighted.append(f1_score(pred4f1[:, 1], pred4f1[:, 0], average='macro'))

        size = min(batch_size, total0 - knownnum)
        if query == 'margin': 
            batch = Make_Selections_score2(size, prediction)
        elif query == 'active':
            batch = Make_Selections_score2(size, prediction)
        elif query == 'entropy':
            batch = Make_Selections_score(1000, prediction)
        else: raise ValueError('unknown')
        affl_matrix.clear()
        mistake = 0
        for p in prediction:
            i = p[0]
            j = p[1]
            label = p[2]
            if ground_truth[i][j] != label:
                mistake += 1
        accuracy.append((total0 - mistake)/total0)
        pred_accuracy.append(1-mistake / len(prediction))
        for item in batch:
            i = item[0]
            j = item[1]
            current_mtx[i][j] = ground_truth[i][j]
        knownnum = CountKnownNum(current_mtx)

    accuracy.append(1.0)
    f1unweighted.append(1.0)
    f1weighted.append(1.0)
    filename = 'onebit_syn_' + str(query) + '_acc.csv'
    result = []
    # result.append([uniqueness, responsiveness, noise])
    result.append(accuracy)
    # result.append(pred_accuracy)
    pd_data = pd.DataFrame(result)
    pd_data.to_csv(filename, index=False, header=False, mode="a")

    result1 = []
    result1.append(f1unweighted)
    pd_data = pd.DataFrame(result1)
    pd_data.to_csv("onebit_syn_" + str(query) + "_unwf1.csv", index=False, header=False, mode="a")
    result2 = []
    result2.append(f1weighted)
    pd_data = pd.DataFrame(result2)
    pd_data.to_csv("onebit_syn_" + str(query) + "_wf1.csv", index=False, header=False, mode="a")

np.random.seed(20)
random.seed(20)
eng = matlab.engine.start_matlab()



# for u in [0.2, 0.4, 0.6, 0.8]:
#     for r in [0.2, 0.4, 0.6, 0.8]:
#         main(u, r, "least_confidence", 0.0)
#         time.sleep(60)

# for u in [0.2, 0.4, 0.6, 0.8]:
#     for r in [0.2, 0.4, 0.6, 0.8]:
#         main(u, r, "least_confidence", 0.1)
#         time.sleep(60)
for noise in [0.0, 0.1, 0.2]:
    for u in [0.2, 0.4, 0.6, 0.8]:
        for r in [0.2, 0.4, 0.6, 0.8]:
            main(u, r, "active", noise)
            time.sleep(60)

for noise in [0.0, 0.1, 0.2]:
    for u in [0.2, 0.4, 0.6, 0.8]:
        for r in [0.2, 0.4, 0.6, 0.8]:
            main(u, r, "random", noise)
            time.sleep(60)





